//  This class handles the Linda primitives.  It is a dictionary for storing
//  keyword-tuples, and can be created with the Scheme function |make-dict|,
//  for local dictionaries.  There is also a global dictionary that is assumed
//  when |Null| is passed for the dictionary parameter.  It is based on the
//  |Hash| class.

#include "Hash.h"
#include "List.h"

//  Each element in our hash table is either a tuple, represented by a list
//  of expressions (in this case, an |Expr| list), or a list of |EvalState|
//  classes (see |Eval.cc|) waiting on a tuple in this dictionary with this
//  keyword (in this case, a real |List|.)

class DictEntry : public Object {
   class Expr *vals ;     // the tuple values
   List *waiting ;        // who is waiting on me?
public:
   DictEntry(class Expr *e) { vals = e ; waiting = new List ; }
   DictEntry(class EvalState *es =0) ;
   ~DictEntry(VOID) ;
   int active(VOID) { return vals != 0 ; }
   void addTask(class EvalState *es) ;
   void scheduleAllTasks(VOID) ;
   void setValue(class Expr *e) { vals = e ; }
   class Expr *val(VOID) { return vals ; }
   virtual const char *isa(VOID) { return "DictEntry" ; }
} ;

class Dict : public Hash {
public:
   Dict(VOID) : () {} ;
   ~Dict(VOID) ;
   DictEntry *find(const char *s) { return (DictEntry *)Hash::find(s) ; }
   void add(const char *s, DictEntry *de) { Hash::add(s, (elem)de) ; }
   virtual const char *isa(VOID) { return "Dict" ; }
} ;

//  We declare our global |Dict| here.

extern Dict *globalDict ;
