//  This is a new attempt at an evaluation routine, but this one avoids
//  using C recursion, so we can trivially implement `threads' or even
//  (glory be!) continuations.  We make the stack explicit.
//
//  Note that this evaluate is no longer a member function; the member
//  function simply calls this function with `this' as an arguement.

#include "EvalState.h"
#define OPTIMIZE

//  We use the C preprocessor to shorten some of the code.

#define popEnv() \
       { scopeenv = popExprPP() ; localenv = getBase() + popInt() ; }
#define pushEnv() \
           { pushInt(localenv - getBase()) ; pushExprPP(scopeenv) ; }
#define popAndGo() { cont = popInt() ; goto cont_dispatch ; }

#define FOREVER (2000000000)

//  This function is called to evaluate on a particular context for a bit.
//  If evaluation doesn't conclude within the duration specified, a 0 is
//  returned; else, a correct value is returned.
//
//  It's a very interesting function.  There are a million labels and
//  goto statements; it's one big state machine.  It's probably not
//  going to be very easy to debug, either, but hey!  It works.

extern Scope *scope ; // our global scope, for adding macros
extern Expr *PrimList(Expr **, int) ;  // a primitive function we call

Expr *EvalState::Evaluate(int duration) {

//  First we declare all of our local state.  This is popped off the
//  stack on entrance, and pushed back on when we exit.
//
//  The two variables |scopeenv| and |localenv| embody our local
//  environment; the former is for variables outside the current
//  closure level, yet not global; the latter is for local variables
//  (arguments to the current closure, or local |let| variables.)
//  The |exp| variable holds the current expression to evaluate
//  and |funklock| points back on the stack to the previous `link'
//  level.  (This is so we can accumulate our procedure arguments on
//  the stack without forgetting where they started.)
//  Finally, |cont| is our continuation argument that indicates where
//  to go after finishing the current evaluation.
//
//  Because this is an explicit control evaluator, all of our forms are
//  `explicit' in this subroutine, so it is a bit long.
//
//  We have yet to do simple optimizations that should speed things up
//  by perhaps a factor of two.

   Expr **scopeenv, **localenv, *exp ;
   int cont ;          // would like to be enum conttype, but that breaks
   int funklock ;

//  Next we declare the local variables used in this procedure.  Some of
//  these may live on the stack at various times, but they do not hold
//  valid data when this function is entered or exited, so they need not
//  be initialized.

   Expr *val, *unev, *fun ;

//  And some local variables for convenience.

   int forever = duration >= FOREVER ;
   int i, j, k, numargs ;
   Expr **args ;
   EvalState *es ;
   Expr *p ;
   Key *key ;
   ScopeEntry *se ;

   popState(&scopeenv, &localenv, &exp, &funklock, &cont) ;

evaluate:

//  Check to see if we are out of time; if so, save the above state and go
//  This is one of the only two exits from this routine; the other is at
//  |RETURN_ME|.

   checkStack(&localenv) ;
   if (duration-- <= 0) {
      if (forever)
         duration = FOREVER ;
      else {
         pushState(scopeenv, localenv, exp, funklock, cont) ;
         return 0 ;  // 0 indicates we aren't quite done yet.
      }
   }
   if (debugging('e')) {
      printf("Evaluating: ") ; exp->display(stdout) ; printf("\n") ;
   }

//  Finally, we get busy.  We key off the type of the expression to figure
//  out what to do.

   switch(exp->etype()) {
   default:
#ifdef DEBUG
      printf("Taking default %d\n", exp->etype()) ;
#endif
      val = exp ;
      break ;
   case EXPR_LOCAL: case EXPR_LOCAL+MAKING_GRAPH:
      val = (localenv[exp->scopeIndex()])->saval() ;
      exp->remove() ;
#ifdef DEBUG
      printf("{%ld}(%d)!\n", localenv, exp->scopeIndex()) ;
      printf(" Found a local of ") ; val->display(stdout) ; printf("\n") ;
#endif
      break ;
   case EXPR_SCOPED: case EXPR_SCOPED+MAKING_GRAPH:
      val = (scopeenv[exp->scopeIndex()])->saval() ;
      exp->remove() ;
#ifdef DEBUG
      printf("Found a scoped of ") ; val->display(stdout) ; printf("\n") ;
#endif
      break ;
   case EXPR_GLOBAL: case EXPR_GLOBAL+MAKING_GRAPH:
      val = exp->globalScopeEntry()->value() ;
      exp->remove() ;
      break ;
   case EXPR_COMPLEX: case EXPR_COMPLEX+MAKING_GRAPH:
      if (exp == Null) {
         val = exp ;
         break ;
      }
      pushInt(cont) ; push(exp) ;      // push <A>
      val = exp->fastcar() ;
#ifdef OPTIMIZE
      if (val->globalp()) {
         unev = exp->fastcdr() ;
         val = val->globalScopeEntry()->value() ;
         goto eval_args_over ;
      } else {
#endif
         pushEnv() ; push(exp->fastcdr()) ;   // push <B>
         exp = val->copy() ;
         cont = EVAL_ARGS ;
         goto evaluate ;
#ifdef OPTIMIZE
      }
#endif
   }

//  At this point, we need to go to where the continuation variable
//  directs us.  This whole system would be significantly faster if we
//  could take the address of a statement in C, and branch directly,
//  rather than having to use a switch statement.
//
//  I prefer putting the default case at the top of switch statements,
//  where it is more easily seen.

cont_dispatch:
#ifdef DEBUG
   printf("Continue is %d (%d)\n", cont, RETURN_ME) ;
#endif
   switch(cont) {
   default:
case ILLEGAL:
      error("! no legal value for continue") ;
case RETURN_ME: case RETURN_ME+MAKING_GRAPH:
#ifdef DEBUG
      printf("Got to return me\n") ;
#endif
      return val ;
case EVAL_ARGS: case EVAL_ARGS+MAKING_GRAPH:
      unev = pop() ; popEnv() ; // pop <B>
eval_args_over:
      if (val->formp()) {
         unev = unev->copy() ; // make sure to delete this as we cdr down it!
         pop()->remove() ;     // pop <A> partially
         cont = val->formv() ;
         goto cont_dispatch ;
      } else if (val->evprocp()) {
lambda_eval:
         push(val) ; pushInt(funklock) ; // push <C>
         funklock = depth() ;
         cont = ACCUMULATE_ARG ;
         goto eval_arg_loop ;
      } else {
         val = pop() ;
         experror("not-a-function in function position", val) ;
         val->remove() ;
         val = Null ;
         popAndGo() ;
      }
case ACCUMULATE_ARG: case ACCUMULATE_ARG+MAKING_GRAPH:
      popEnv() ; unev = pop() ; // pop <D>
      cont = ACCUMULATE_ARG ;
push_arg_val:
      push(val) ;               // accumulate this argument
eval_arg_loop:
      if (unev != Null) {
         val = unev->fastcar() ;
         unev = unev->fastcdr() ;
#ifdef OPTIMIZE
         switch (val->etype()) {
case EXPR_SCOPED: case EXPR_SCOPED+MAKING_GRAPH:
            val = (scopeenv[val->scopeIndex()])->saval() ;
            goto push_arg_val ;
case EXPR_GLOBAL: case EXPR_GLOBAL+MAKING_GRAPH:
            val = val->globalScopeEntry()->value() ;
            goto push_arg_val ;
case EXPR_LOCAL: case EXPR_LOCAL+MAKING_GRAPH:
            val = (localenv[val->scopeIndex()])->saval() ;
            goto push_arg_val ;
default:
            val = val->copy() ;
            goto push_arg_val ;
case EXPR_COMPLEX: case EXPR_COMPLEX+MAKING_GRAPH:
#endif
            exp = val->copy() ;
            push(unev) ; pushEnv() ;  // push <D>
            goto evaluate ;
#ifdef OPTIMIZE
         }
#endif
      }
      numargs = depth() - funklock ;
      setSP(funklock) ;
      args = getSP() ;
      funklock = popInt() ; fun = pop() ; // pop <C>
      exp = pop() ; // pop part of <A>
      if (fun->lambdap()) {

//  Here we need to create a new environment, from the arguments on the
//  stack and the lambda context.  Can we do it?
//
//  We still need to make sure that we don't delete the scope environment
//  out from under us, which might occur if the lambda we are executing
//  gets deleted while we are executing it.  (The body is protected because
//  we copy it before we execute it, since execution deletes the thing
//  that is being executed.)
//
//  This is done by pushing `fun' on the local stack along with the local
//  environment.  When that local environment is popped, then the scoped
//  environment can (will) be deleted.

         cont = popInt() ;
         if (cont == POP_LOCALS) {

//  Tail recursive call!  Pop old stack before we create new one.
//  This can be `optimized' to use the old bindings, but we want to
//  get it working before we get it working fast.

            for (i=popInt(); i>0; i--)
               pop()->remove() ;
         } else {
            pushInt(cont) ;
            cont = POP_LOCALS ;
         }
         k = fun->lambdaMax() - fun->lambdaLow() ;
#ifdef DEBUG
 printf("Allocating local space for %d\n", k) ;
#endif
         checkSpace(k, &localenv) ;
         localenv = getSP() ;
         scopeenv = fun->lambdaSA() ;
         j = fun->lambdaArity() ;
         if (j > k)
            error("! Something's very wrong.\n") ;
         val = fun->lambdaArgs() ;
         i = 0 ;
         j = 0 ;
         while (val != Null) {
            if (val->consp()) {
               if (i >= numargs)
                  push(Null) ;
               else
                  push(new Expr(val->fastcar()->boundScopeEntry()->
                                           typeCast(*args++), EXPR_BIND)) ;
               val = val->cdr() ;
               i++ ;
               j++ ;
            } else {
               val = Null ;
               for (;i<numargs; i++)
                  val = new Expr(args[numargs-i-1], val) ;
               push(new Expr(val, EXPR_BIND)) ;
               j++ ;
               val = Null ;
            }
         }
         if (i != numargs) {
            experror("wrong number of arguments to function", exp) ;
            while (i < numargs) {
               (*args++)->remove() ;
               i++ ;
            }
         }
         for (; j<k; j++)
            push(Null) ;
         push(fun) ;
         pushInt(k+1) ;
         exp->remove() ;
         exp = fun->lambdaExpr()->copy() ;
         goto evaluate ;
      } else if (fun->evformp()) {  // okay, a form that evaluates its args
         cont = fun->formv() ;
         goto cont_dispatch ;

//  Otherwise we simply have a primitive call.  We make the call and then
//  delete the arguments.

      } else {
         cont = popInt() ;
         if (hwsg && fun->hardprimp()) {
            val = Null ;
            for (i=0; i<numargs; i++)
               if (args[i]->nodep()) {
                  val = new Expr(hwsg->callPrimGraph(exp->car(), args,
                                                                numargs)) ;
                  break ;
               }
            if (val == Null)
               val = fun->callPrim(args, numargs) ;
         } else
            val = fun->callPrim(args, numargs) ;
         exp->remove() ;
         fun->remove() ;
         for (i=0; i<numargs; i++)
            (*args++)->remove() ;
         goto cont_dispatch ;
      }
 case POP_LOCALS: case POP_LOCALS+MAKING_GRAPH:
      for (i=popInt(); i>0; i--)
         pop()->remove() ;
      popAndGo() ;
 case QUOTE_FORM: case QUOTE_FORM+MAKING_GRAPH:
      val = unev->fastcar()->copy() ;
      goto cleanAndGo ;

//  To evaluate a quasiquote form, we use the stack.  This must
//  work recursively, and we have to do everything within the
//  context of our main evaluation routine.  Macros are a mess, aren't they?
//
//  The neat thing about the following code is it is sort of like a miniature
//  version of the entire interpreter, in the way that it squeezes recursion
//  onto the stack.

 case QQUOTE_FORM: case QQUOTE_FORM+MAKING_GRAPH:
      i = 1 ;      // this is the current `level'
      cont = 0 ;   // do next:  return
      exp = unev->car()->copy() ; // only look at the car
      unev->remove() ;
      while (1) {
         if (debugging('q')) {
            printf("Level = %d checking ", i) ;
            exp->display(stdout) ;
            printf("\n") ;
         }
         if (exp->consp() && exp != Null) {
            unev = exp->car() ;
            if (unev->globalp() &&
                (p=unev->globalScopeEntry()->valueNoCopy())->formp() &&
                p->formv()==UNQUOTE_FORM) {
               if (debugging('q'))
                  printf("Hit!!\n") ;
               i-- ;
               if (i <= 0) {
                  pushInt(cont) ;
                  pushEnv() ;
                  cont = QQUOTE_RETURN ;
                  exp = exp->cdr()->car()->copy() ;
                  goto evaluate ;
 case QQUOTE_RETURN: case QQUOTE_RETURN+MAKING_GRAPH:
                  popEnv() ;
                  cont = popInt() ;
                  i = 0 ;
                  goto gotval ;
               }
            }
            pushInt(cont) ;
            pushInt(i) ;
            push(exp->car()->copy()) ;
            exp = exp->cdr()->copy() ;
            cont = 1 ; // join next
            continue ;
         } else {
            val = exp->copy() ;
         }
       gotval:
         if (cont == 0)
            break ;
         else if (cont == 1) {
            exp = pop() ;
            i = popInt() ;
            cont = 2 ;
            push(val) ;
         } else if (cont == 2) {
            val = new Expr(val, pop()) ;
            cont = popInt() ;
            goto gotval ;
         } else
            error("! continue has bad value during quasiquote") ;
      }
      popAndGo() ;
 case UNQUOTES_FORM: case UNQUOTES_FORM+MAKING_GRAPH:
 case UNQUOTE_FORM: case UNQUOTE_FORM+MAKING_GRAPH:
      error("Use of unquote outside of quasiquote") ;
      goto cleanAndReturnNull ;
 case SIM_FORM: case SIM_FORM+MAKING_GRAPH:
 case CONS_FORM: case CONS_FORM+MAKING_GRAPH:
 case MODEL_FORM: case MODEL_FORM+MAKING_GRAPH:
 case STIM_FORM: case STIM_FORM+MAKING_GRAPH:
      push((Expr *)hwsg) ;
      pushInt(STIM_FORM_2) ;
      hwsg = 0 ;
 case BEGIN_FORM: case BEGIN_FORM+MAKING_GRAPH:
 begin_form: 
      if (unev) {
         cont = BEGIN_FORM_3 ;
         push(unev) ;          // push <G>
         goto begin_form_2 ;
      } else {
         val = Null ;
         popAndGo() ;
      }
 case BEGIN_FORM_3: case BEGIN_FORM_3+MAKING_GRAPH:
      popEnv() ; unev = pop() ; // pop <H>
      val->remove() ;
 begin_form_2:
      exp = unev->fastcar()->copy() ;
      unev = unev->fastcdr() ;
      if (unev != Null) {
         push(unev) ; pushEnv() ; // push <H>
      } else {
         unev = pop() ; // pop <G>
         cont = popInt() ;
         unev->remove() ;
      }
      goto evaluate ;
 case STIM_FORM_2: case STIM_FORM_2+MAKING_GRAPH:
      hwsg = (HWSG *)pop() ;
      popAndGo() ;
 case IF_FORM: case IF_FORM+MAKING_GRAPH:
      exp = unev->fastcar()->copy() ;
      push(unev->fastcdr()->copy()) ; pushEnv() ; // push <I>
      unev->remove() ;
      cont = IF_FORM_2 ;
      goto evaluate ;
 case IF_FORM_2: case IF_FORM_2+MAKING_GRAPH:
      popEnv() ; unev = pop() ; // pop <I>
      if (hwsg && val->nodep()) {
         push(val) ;
         push(unev->fastcdr()->copy()) ;
         pushEnv() ;
         cont = IF_FORM_3 ;
         exp = unev->fastcar()->copy() ;
         unev->remove() ;
         goto evaluate ;
      }
      cont = popInt() ;
      exp = (val != Null ? unev : unev->fastcdr())->fastcar()->copy() ;
      unev->remove() ;
      val->remove() ;
      goto evaluate ;
 case IF_FORM_3: case IF_FORM_3+MAKING_GRAPH:
      popEnv() ;
      unev = pop() ;
      push(val) ;
      cont = IF_FORM_4 ;
      exp = unev->fastcar()->copy() ;
      unev->remove() ;
      pushEnv() ;
      goto evaluate ;
 case IF_FORM_4: case IF_FORM_4+MAKING_GRAPH:
      popEnv() ;
      exp = pop() ;
      unev = pop() ;
      val = new Expr(hwsg->addIfGraph(unev, exp, val)) ;
      popAndGo() ;
 case DEFINE_FORM: case DEFINE_FORM+MAKING_GRAPH:
 case SET_FORM: case SET_FORM+MAKING_GRAPH:
 case DEF_MAC_FORM: case DEF_MAC_FORM+MAKING_GRAPH:
      push(unev) ; pushEnv() ; // push <L>
      exp = unev->fastcdr()->fastcar()->copy() ;
      cont++ ;
      goto evaluate ;
 case DEFINE_FORM_2: case DEFINE_FORM_2+MAKING_GRAPH:
 case SET_FORM_2: case SET_FORM_2+MAKING_GRAPH:
      popEnv() ; unev = pop() ; // pop <L>
      unev->fastcar()->setBang(scopeenv, localenv, val) ;
      val = unev->fastcar()->copy() ;
      goto cleanAndGo ;
 case LAMBDA_FORM: case LAMBDA_FORM+MAKING_GRAPH:
      val = unev->lambdaToClosure(scopeenv, localenv) ;
      goto cleanAndGo ;
 case LET_FORM: case LET_FORM+MAKING_GRAPH:
      push(unev) ;
      push(unev->fastcar()->fastcar()->fastcdr()) ;
      unev = unev->fastcar()->fastcdr() ;
      cont = LET_FORM_2 ;
      goto let_form_3 ;
 case LET_FORM_2: case LET_FORM_2+MAKING_GRAPH:
      popEnv() ; unev = pop() ;
      exp = pop() ;
      exp->fastcar()->setBang(scopeenv, localenv, val, 1) ;
      push(exp->fastcdr()) ;
let_form_3:
      if (unev != Null) {
         exp = unev->fastcar()->copy() ;
         unev = unev->fastcdr() ;
         push(unev) ; pushEnv() ;
      } else {
         pop() ;
         unev = pop() ;
         exp = unev->fastcdr()->fastcar()->copy() ;
         unev->remove() ;
         cont = popInt() ;
      }
      goto evaluate ;
 case NAMED_LET_FORM: case NAMED_LET_FORM+MAKING_GRAPH:
      push(unev) ;
      val = unev->namedLetToClosure(scopeenv, localenv) ;
      (val->lambdaSA())[val->lambdaLow()-1]->setcar(val) ;
      unev = unev->fastcar()->fastcdr()->fastcdr()->fastcdr()->fastcdr() ;
      goto lambda_eval ;
 case DISPLAY_TYPE_FORM: case DISPLAY_TYPE_FORM+MAKING_GRAPH:
      se = unev->fastcar()->scopeEntry() ;
      if (se)
         se->eType()->display() ;
      else
         printf("No type\n") ;
 cleanAndReturnNull:
      val = Null ;
 cleanAndGo:
      unev->remove() ;
      popAndGo() ;
 case DEF_MAC_FORM_2: case DEF_MAC_FORM_2+MAKING_GRAPH:
      popEnv() ; unev = pop() ; // pop <L>
      unev->fastcar()->setBang(scopeenv, localenv, val) ;
      unev->fastcar()->scopeEntry()->setIsMacro() ;
      val = unev->fastcar()->copy() ;
      goto cleanAndGo ;
 case SPAWN_FORM: case SPAWN_FORM+MAKING_GRAPH:
      for (i=0; i < numargs; i++) {
         es = new EvalState(this, args[i], WAIT) ;
         es->schedule() ;
      }
      pushState(scopeenv, localenv, Null, funklock, SPAWN_DONE) ;
      setWaitCount(i) ;
      exp->remove() ;
      return(0) ;
 case WAIT: case WAIT+MAKING_GRAPH:
      es = popES() ;
      val->remove() ;
      es->decrementWaitCount() ;
      this->unschedule() ;
      delete this ;
      return 0 ;

//  Now we handle the Linda forms.  At this point, |numargs|, |args|,
//  |exp| are all valid, and the next |cont| is on the stack.
//
//  The reason these cannot be functions is because they manipulate the
//  local state in order to schedule.

 case IN_FORM: case IN_FORM+MAKING_GRAPH:
      if (numargs != 1) {
         error("need only one argument for IN function") ;
         goto linda_error ;
      }
      p = *args ;
      if (! p->keyp()) {
         error("first arg to in function must be a key") ;
         goto linda_error ;
      } else
         key = p->keyv() ;
      exp->remove() ;
      if (key->voteIn() > 0) { // best case: grab an |OUT| and pass it in
         val = key->getKeyVal() ;
         key->setKeyVal(0) ;   // necessary to prevent double deletion
         es = key->getFirstTask() ;
         if (es) {
            key->setKeyVal(es->getSlot()) ;
            es->schedule() ;
         }
         p->remove() ;
         popAndGo() ;
      } else
         goto linda_unschedule ;
 case IN_ASSIGN: case IN_ASSIGN+MAKING_GRAPH:
      val = this->getSlot() ;
      popAndGo() ;
 case OUT_FORM: case OUT_FORM+MAKING_GRAPH:
      if (numargs < 1 || numargs > 2) {
         error("need one or two arguments for out") ;
         goto linda_error ;
      }
      p = *args ;
      if (! p->keyp()) {
         error("first arg to linda function must be a key") ;
         goto linda_error ;
      } else
         key = p->keyv() ;
      exp->remove() ;
      val = (numargs > 1) ? args[1] : Null ;
      i = key->voteOut() ;
      if (i==0) {
         key->setKeyVal(val) ;
      } else if (i > 0) {
         this->setSlot(val) ;
         goto linda_unschedule ;
      } else {
         es = key->getFirstTask() ;
         es->setSlot(val) ;
         es->schedule() ;
      }
      p->remove() ;
 case OUT_ASSIGN: case OUT_ASSIGN+MAKING_GRAPH:
 case SPAWN_DONE: case SPAWN_DONE+MAKING_GRAPH:
      val = Null ;
      popAndGo() ;
 case TIME_FORM: case TIME_FORM+MAKING_GRAPH:
      push(unev) ; pushEnv() ; // push <M>
      exp = unev->fastcar()->copy() ;
      cont = TIME_FORM_2 ;
      pushInt(ourtime()) ;
      goto evaluate ;
 case TIME_FORM_2: case TIME_FORM_2+MAKING_GRAPH:
      printf("Time: %d\n", ourtime() - popInt()) ;
      popEnv() ; unev = pop() ; // pop <M>
      goto cleanAndGo ;
 case MAKE_GRAPH_FORM: case MAKE_GRAPH_FORM+MAKING_GRAPH:
      if (numargs != 1) {
         error("need only one argument for MAKEGRAPH function") ;
         goto linda_error ;
      }
      p = *args ;
      if (! p->lambdap()) {
         error("first arg to in function must be a lambda") ;
         goto linda_error ;
      }
      if (hwsg) {
         error("already building graph") ;
         goto linda_error ;
      }
      hwsg = new HWSG() ;
      exp->remove() ;
      exp = hwsg->startLambda(p) ;
      duration = FOREVER ;
      cont = MAKE_GRAPH_RETURN ;
      goto evaluate ;
 case MAKE_GRAPH_RETURN: case MAKE_GRAPH_RETURN+MAKING_GRAPH:
      val->remove() ;
      val = new Expr(hwsg) ;
      hwsg = 0 ;
      popAndGo() ;
   }

//  If we block for some reason on a Linda call, we restore the stack
//  to its state before we `popped' the args, and exit.  Note that this
//  causes a little extra bit of code to be called each time we reenter,
//  but since we usually only reenter when we need to (unless a lot of
//  people are waiting), this is no big deal.

linda_unschedule:
   i = ((cont == IN_FORM) ? IN_ASSIGN : OUT_ASSIGN) ;
   pushState(scopeenv, localenv, Null, funklock, i) ;
   this->unschedule() ;
   key->addTask(this) ;
   p->remove() ;
   return 0 ;
linda_error:
   cont = popInt() ;
   for (i=0; i<numargs; i++)
      args[i]->remove() ;
   exp->remove() ;
   val = Null ;
   goto cont_dispatch ;
}

//  Finally, the function that completely evaluates something.
//  (Our scheduler to follow.)

#define TIMESLICE (50)

Expr *Expr::eval(int locals) {
   Expr *ret ;
   List readyQ ;
   EvalState localEvalState(this, locals, &readyQ) ;
   localEvalState.schedule() ;
   while (1) {
      EvalState *es = (EvalState *)readyQ.firstEl() ;
      if (debugging('e'))
         printf("Tick %ld\n", es) ;
      if (es == 0) {
         error("deadlock situation") ;
         return Null ;
      }
      ret = es->Evaluate(TIMESLICE) ;
      if (ret) {
         es->unschedule() ;
         if (readyQ.length() > 0)
            error("still processes on the queue") ;
         break ;
      }
      if (readyQ.length() > 1 && readyQ.firstEl() == es) {
         es->unschedule() ;
         es->schedule() ;
      }
   }
   return ret ;
}
