//  This is our |EvalState| class, that stores the evaluation context for our
//  tasking.

 // #define DEBUG ABSOLUTELY

#include "HWSG.h"
#include "Key.h"

//  These are all the form constants needed for our
//  explicit control evaluator.  The identifiers ending in just
//  |_FORM| are also used in the symbol table, to tell where to branch
//  to when that form is found.

enum conttype { ILLEGAL, ACCUMULATE_ARG, EVAL_ARGS, RETURN_ME, POP_LOCALS,
                IF_FORM, IF_FORM_2, IF_FORM_3, IF_FORM_4,
                BEGIN_FORM,
                BEGIN_FORM_3, DEFINE_FORM, DEFINE_FORM_2, SET_FORM,
                SET_FORM_2, LAMBDA_FORM, LET_FORM, LET_FORM_2,
                NAMED_LET_FORM, TIME_FORM, TIME_FORM_2, DISPLAY_TYPE_FORM,
                DEF_MAC_FORM, DEF_MAC_FORM_2,
                QUOTE_FORM, QQUOTE_FORM, UNQUOTE_FORM, UNQUOTES_FORM,
                QQUOTE_RETURN,
                IN_FORM, IN_ASSIGN, OUT_FORM,
                OUT_ASSIGN, SPAWN_FORM, SPAWN_DONE, WAIT, MAKE_GRAPH_FORM,
                MAKE_GRAPH_RETURN, MODEL_FORM, STIM_FORM, STIM_FORM_2,
                SIM_FORM,
                CONS_FORM, MAKING_GRAPH } ;

//  Before we get into the actual evaluation procedure, we need to define
//  a class that contains all of the evaluation state, and various
//  procedures to manipulate this state.  Since this class is only used
//  here, we simply inline the declaration.  (This tends to make this
//  particular file rather long, but there is a lot going on here, and
//  all of this really belongs together.)

class EvalState : public Object {
   Expr **estack, **esp, **esend ; // the subproblem stack
   int essize ;                    // how big the stack is
   int waitcount ;                 // how many are we waiting on
   link eslink ;                   // our list link
   List *readyQ ;                  // the list of processes
   Expr *valslot ;                 // place to put a value
public:
   EvalState(Expr *e, int locals, List *rQ) ;
   EvalState(EvalState *es, Expr *e, int cont) ;
   ~EvalState(VOID) ;
   Expr *Evaluate(int duration) ;
   void push(Expr *e) { *esp++ = e ; }
   void pushExprPP(Expr **e) { *esp++ = (Expr *)e ; }
   void pushInt(int e) { *esp++ = (Expr *)e ; }
   Expr *pop(VOID) { return *--esp ; }
   Expr **popExprPP(VOID) { return (Expr **)*--esp ; }
   int popInt(VOID) { return (int)*--esp ; }
   void pushES(EvalState *e) { *esp++ = (Expr *)e ; }
   EvalState *popES(VOID) { return (EvalState *)*--esp ; }
   int empty(VOID) { return esp < estack ; }
   int depth(VOID) { return esp - estack ; }
   void setSP(int depth) { esp = estack + depth ; }
   void setSP(Expr **e) { esp = e ; }
   Expr **getSP(VOID) { return esp ; }
   Expr **getBase(VOID) { return estack ; }
   void growStack(Expr ***locptr) ;
   void checkStack(Expr ***locptr) { if (esp >= esend) growStack(locptr) ; }
   void checkSpace(int n, Expr ***locptr)
                               { if (esp + n >= esend) growStack(locptr) ; }
   void pushState(Expr **scopeenv, Expr **locenv, Expr *exp, int funklock,
                                           int cont) ;
   void popState(Expr ***scopeenv, Expr ***locenv, Expr **exp, int *funklock,
                                           int *cont) ;
   void schedule(VOID) ;
   void unschedule(VOID) ;
   void setWaitCount(int n) { waitcount = n ; if (waitcount) unschedule() ; }
   void decrementWaitCount(VOID) { if (--waitcount == 0) schedule() ; }
   link *ESlink(VOID) { return &eslink ; }
   virtual const char *isa(VOID) { return "EvalState" ; }
   Expr *getSlot(VOID) { return valslot ; }
   void setSlot(Expr *e) { valslot = e ; }
} ;
