//  This class implements a simple hash table for strings.  As in the
//  |list| class, we deal with arbitrary pointers (|void *|), defined
//  in |misc.h|.
//
//  The hash table grows dynamically to keep the average length of
//  the hash chains under two.  Each time the hash table gets a bit
//  too long, it is expanded by a factor of slightly more than three,
//  and the entire table is rehashed.

#ifndef HASH_CLASS
#define HASH_CLASS
#include "Object.h"

//  First, we need a type for our hash links.  This is private.  Not much
//  to this.

struct _hashlink {
   struct _hashlink *next ;
   elem e ;
   char name[1] ;
} ;

//  The accessible functions are creation, deletion, finding an element,
//  adding a new string/element association, and removing an association.
//  There is also a |nameFind| routine supplied that is useful when you
//  wish to use the string pointer as a unique string; this is allocated
//  when an element is added.  This way you can use the hash table to
//  `allocate' string space for your identifiers as well, or you can
//  point a `name' field of some structure here.

class Hash : public Object {
   friend class Scope ;   // so we can clean up at the end
   friend class Dict ;    // ditto
   int hashPrime ;   // how big is our hash table?
   struct _hashlink **hashtable ; // our actual hash table
   int hashVal(const char *s) ;
   int count ;       // how many elements are in our hash table?
   void rehash(VOID) ;   // reallocate; we got too full.
public:
   Hash(int numbuckets =7) ;
   ~Hash(VOID) ;
   elem find(const char *s) ;
   char *nameFind(const char *s) ;
   elem add(const char *s, elem e) ;
   elem remove(const char *s) ;
   virtual const char *isa(VOID) { return "Hash" ; }
} ;
#endif
