//  Our `implementation' of the |List| class is fairly trivial.

#include "List.h"

//  Our |insert| and |append| functions are fairly standard linked list
//  operations.

void link::insert(link *what) {
   what->next = this ; what->prev = this->prev ;
   this->prev->next = what ; this->prev = what ;
}

void link::append(link *what) {
   what->prev = this ; what->next = this->next ;
   this->next->prev = what ; this->next = what ;
}

//  Finally, our list functions.  Some of these could be inline, but
//  the code gets awfully big awfully quick.

link *List::first(VOID) {
   return ((count) ? (head->next) : 0) ;
}

elem List::takeFirstEl(VOID) {
   if (count) {
      link *l = head->next ;
      l->removeNoDealloc() ;
      return l->e ;
   } else
      return 0 ;
}

link *List::last(VOID) {
   return ((count) ? (head->prev) : 0) ;
}

link *List::next(link *what) {
   return ((what->next == head) ? 0 : what->next) ;
}

link *List::prev(link *what) {
   return ((what->prev == head) ? 0 : what->prev) ;
}

elem List::el(link *what) {
   return ((what == 0 || what == head) ? 0 : what->e) ;
}

//  Our constructor and destructor are also nontrivial enough to deserve
//  functions.

List::List(elem hdata=0) {
   head = new link(hdata) ;
   head->next = head ;
   head->prev = head ;
   count = 0 ;
}

List::~List() {
   while (head->next != head)
      head->next->remove() ;
   delete head ;
}

//  This function `finds' the first occurence of an element in the list.
//  Note that only the pointer is compared, making it a very shallow
//  find.

link *List::find(elem el) {
   ListIterator li(this) ;
   elem e ;
   while (e=li())
      if (e==el)
         return(li.curlink()) ;
   return(0) ;
}

//  This function finds a particular element and removes it from the list
//  if it was found.  If the element is found (and removed) this function
//  returns |1|, otherwise it returns |0|.

int List::findAndRemove(elem el) {
   ListIterator li(this) ;
   elem e ;
   while (e=li())
      if (e==el) {
         li.removeForward() ;
         return(1) ;
      }
   return(0) ;
}

//  Our general iteration operators.

elem ListIterator::operator()() {
   return (ourlist->el(curplace=curplace->next)) ;
}

elem ListIterator::next() {
   return (ourlist->el(curplace=curplace->next)) ;
}

elem ListIterator::prev() {
   return (ourlist->el(curplace=curplace->prev)) ;
}

//  Now we define the |remove_forward()| and |remove_backward()| functions
//  of our iterator.  These each return |0| if there is no current point
//  or if there is no next (or previous) element.
//
//  In a typical loop such as |while (e=next())|, if you want to delete
//  an element without messing up the logic of the loop, you should use
//  the `wrong' one (ie, |remove_backward| in this case) so the next
//  iteration catches the correct next element.

elem ListIterator::removeForward() {
   link *t = curplace ;
   if (t) {
      elem e = t->prev ;
      t->remove() ;
      ourlist->count-- ;
      return(e) ;
   } else
      return(0) ;
}

elem ListIterator::removeBackward() {
   link *t = curplace ;
   if (t) {
      elem e = t->next ;
      t->remove() ;
      ourlist->count-- ;
      return(e) ;
   } else
      return(0) ;
}
