//  We need a couple of functions for our Scope class.

#include "Scope.h"

//  Our |pop| member has to deallocate itself.

Scope *Scope::pop() {
   Scope *next = this->uplevel() ;
   delete this ;
   return next ;
}

//  Our destructor simply deallocates what was allocated above.

Scope::~Scope() {
   int i ;
   struct _hashlink *hl, *n ;

   for (i=0; i<hashPrime; i++)
      for (hl = hashtable[i]; hl; hl = n) {
         n = hl->next ;
         delete (ScopeEntry *)hl->e ;
      }
}

//  Our |findInTree| finds the name, wherever it resides.  It simply
//  recursively searches the scope pointers up.
//
//  At some point we can make this more efficient by passing in the
//  hash value, but at the moment we are more concerned with making
//  things work.

ScopeEntry *Scope::findInTree(const char *s) {
   ScopeEntry *t = find(s) ;
   if (t)
      return(t) ;
   else if (prevscope == 0)
      return(0) ;
   else
      return(prevscope->findInTree(s)) ;
}

//  Whenever we add something to a scope, we set the |id| value to the
//  pointer that's allocated in the hash list.

void Scope::add(const char *s, scopeentrytype t, Expr *v, class Type *ty =0) {
   ScopeEntry *se = new ScopeEntry(t, v) ;
   delete (ScopeEntry *)Hash::add(s, (elem)se) ;
   se->setId(nameFind(s)) ;
   se->setType(ty) ;
}

void Scope::add(const char *s, Expr *v, class Type *t =0) {
   Scope::add(s, SCOPE_EXPR, v, t) ;
}

//  Our problem with |addBound| is that it creates a |ScopeEntry| that
//  points to an |Expr| that points back to the |ScopeEntry|.  Not
//  difficult, but we want to make sure we get it right.
//
//  This code defaults to creating an |EXPR_SCOPED|, because they at
//  least will maintain their depth from the outside as an index.
//  The |EXPR_LOCAL| will not, so they need to be special cased in
//  |bind.cc|.  This documentation stinks, sorry.

void Scope::addBound(const char *s, int level, class Type *t =0) {
   ScopeEntry *se = new ScopeEntry(SCOPE_EXPR, (Expr *)0) ;
   Expr *v = new Expr(se, level) ;
   se->setValue(v) ;
   se->setType(t) ;
   delete (ScopeEntry *)Hash::add(s, (elem)se) ;
   se->setId(nameFind(s)) ;
}

//  For typed declarations, we call this routine.  It adds the
//  appropriate symbol table entry and returns a type pointer.
//  First, it attempts to determine if it already exists; if it
//  does, it just hacks the old version.

const char *Scope::declare(class Expr *te, int level = -1) {
   Type *t = new Type(te) ;
   const char *r = 0 ;
   if (t==0)
      error("! can't happen: Type::Type returned 0") ;
   if ((r=t->typeId())==0)
      error("declaration requires an identifier") ;
   else {
      ScopeEntry *se = find(r) ;
      if (se) {
         delete se->eType() ;
         se->setType(t) ;
      } else {
         if (level >= 0)
            addBound(r, level, t) ;
         else
            add(r, Null, t) ;
      }
   }
   return r ;
}
