//  This subclass of |Hash| implements a simple symbol table.

#include "Hash.h"
#include "Expr.h" // for evalftype
#include "Type.h" // so inline deletion works

//  Our entries.  The types we allow are

enum scopeentrytype { SCOPE_MACRO, SCOPE_EXPR } ;

//  We need a function that does nothing other than return |Null| for the
//  times a function is undefined.  This function is defined in |Define.cp|.

extern evalftype nullreturn ;

//  Our |ScopeEntry| type contains the actual information.  We have a
//  |Expr *| for the first three and a pointer to a function that takes
//  a |Expr *| and returns a |Expr *| for the last two.

class ScopeEntry : public Object {
   short type, trace ;
   Expr *val ;
   const char *idptr ;
   Type *etype ;
public:
   ScopeEntry(scopeentrytype t, Expr *v) { type = t ; val = v ;
                      idptr = 0 ; etype = 0 ; trace = 0 ; }
   ~ScopeEntry(VOID) { val->remove() ; if (etype) delete etype ; }
   Expr *value(VOID) { return val->copy() ; }
   Expr *valueNoCopy(VOID) { return val ; }
   Type *eType(VOID) { return etype ; }
   Expr *typeCast(Expr *s) { return eType()->typeCast(s) ; }
   simpleMacrop(VOID) { return (type==SCOPE_MACRO && ! val->lambdap()) ; }
   complexMacrop(VOID) { return (type==SCOPE_MACRO && val->lambdap()) ; }
   exprp(VOID) { return type==SCOPE_EXPR ; }
   void setValue(Expr *v) { val = v ; }
   void setId(const char *idp) { idptr = idp ; }
   void setType(Type *t) { etype = t ; }
   void setIsMacro(VOID) { type = SCOPE_MACRO ; }
   const char *id(VOID) { return idptr ; }
   virtual const char *isa(VOID) { return "ScopeEntry" ; }
   int traced(VOID) { return trace ; }
   void setTraced(int n) { trace = n ; }
} ;

//  Our actual |Scope| class then becomes just a few casts from |Hash|.

class Scope : public Hash {
   Scope *prevscope ; // what's the next outer scope (if any)
public:
   Scope(Scope *currscope =0) : () { prevscope = currscope ; } ;
   ~Scope(VOID) ;
   ScopeEntry *find(const char *s)
      { return (ScopeEntry *)Hash::find(s) ; }
   ScopeEntry *findInTree(const char *s) ;
   void add(const char *, scopeentrytype, Expr *, Type *t =0) ;
   void add(const char *, Expr *, Type *t =0) ;
   void addBound(const char *, int, Type *t =0) ;
   virtual const char *isa(VOID) { return "Scope" ; }
   const char *declare(Expr *, int level = -1) ;
   Scope *push(VOID) { return new Scope(this) ; }
   Scope *pop(VOID) ;
   Scope *uplevel(VOID) { return prevscope ; }
} ;

//  We need a function to set up our primitives at the very beginning.

extern void initprimitives(Scope *) ;
