//  The implementation of our |Text| class has only to implement the
//  read and write operations on files, and the clear and
//  deallocate functions.

#include "Text.h"

//  To read text from a file, we try and open the file, and return
//  0 if that fails.  Otherwise, we call the overloaded function,
//  close the file, and return.

int Text::readTextFromFile(char *s) {
   FILE *f = fopen(s, "r") ;
   if (f==NULL)
      return(0) ;
   int ret = readTextFromFile(f) ;
   fclose(f) ;
   return(ret) ;
}

//  When we're reading from the file, we use a buffer on the stack.
//  We have to kill the newlines as we read them.
//  We always succeed at the moment.

#define BUFFSIZE (1024)

int Text::readTextFromFile(FILE *f) {
   char tmp[BUFFSIZE] ;

   clear() ;
   while (fgets(tmp, BUFFSIZE, f)) {
      tmp[strlen(tmp)-1] = 0 ; // kill terminating newline
      appendLine(tmp) ;
   }
   return(1) ;
}

//  Writing text to a file from a name works exactly the same as
//  the function above.

int Text::writeTextToFile(char *s) {
   FILE *f = fopen(s, "w") ;
   if (f==NULL)
      return(0) ;
   int ret = writeTextToFile(f) ;
   fclose(f) ;
   return(ret) ;
}

//  The actual write operation is simple.  Still don't check error
//  returns.

int Text::writeTextToFile(FILE *f) {
   Line *curline = firstLine() ;
   while (curline) {
      fputs(curline->data(), f) ;
      fputc(10, f) ;
      curline = nextLine(curline) ;
   }
   return(1) ;
}

//  To clear our |Text|, we effectively delete the current list and
//  allocate a new one.

void Text::clear() {
   deleteAListOf(Line, lines) ;
   delete lines ;
   lines = new List ;
}
