//  This file declares the |Type| class.
//  The types we support are:
//
//     signed value of $m$ bits, $32 \leq m < 0$
//     unsigned value of $m$ bits, $32 \leq m < 0$
//     signed wire of $m$ bits, $32 \leq m < 0$
//     unsigned wire of $m$ bits, $32 \leq m < 0$
//     closure, with a certain arity, each argument with a type,
//        and with the closure returning a type.
//     no return, a type indicating no return from this closure.
//     unspecified, indicating no known type is returned.

#include "Object.h"

enum typetype { TYPE_VALUE, TYPE_WIRE, TYPE_CLOSURE, TYPE_NORETURN,
                TYPE_KEY, TYPE_UNSPECIFIED } ;

//  These are the flags that are possible.

#define TYPEF_UNSIGNED (1)
#define TYPEF_INPUT_ONLY (2)
#define TYPEF_OUTPUT_ONLY (4)

class Type : public Object {
   short width ;    // how wide the type is in bits, or the arity+1 if a closure
   short type ;     // the type of this type (from the above enumeration)
   short flags ;    // whether the type is signed or no
   const char *id ; // the `id' associated with this type, if one <kludge>
   class Type **closuretypes ;  // if a closure, the list of types
   long mask, smask ; // for quicker evaluation, we build the masks early.
public:
   Type(class Expr *) ;
   ~Type(VOID) ;
   class Expr *typeCast(Expr *e) ;
   const char *typeId(VOID) { return id ; }
   int typeWidth(VOID) { return width ; }
   int fixVal(int n) ;
   int wirep(VOID) { return type == TYPE_WIRE ; }
   virtual const char *isa(VOID) { return "Type" ; }
   void display(int level =0) ;
} ;
