/*  This function defines the four system-dependent windowing functions. */

#include <stdio.h>

#ifdef CPLUS
#define VOID void
#else
#define VOID
#endif

extern char *progname ;

#ifdef X
#ifdef CPLUS
#include <std.h>
/*  the X includes use these as names */
#define new was_new
#define class was_class
#define XOpenDisplay not_XOpenDisplay
#define XCreateWindow not_XCreateWindow
#define XLoadQueryFont not_XLoadQueryFont
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#undef XOpenDisplay
#undef XCreateWindow
#undef XLoadQueryFont
/* now *we* define all necessary functions. */
extern "C" Display *XOpenDisplay(char *) ;
extern "C" void XDestroyWindow(Display *, Window) ;
extern "C" Window XCreateWindow(Display *, Window, int, int, int, int, int,
      int, int, Visual *, int, XSetWindowAttributes *) ;
extern "C" void XSetStandardProperties(Display *, Window, char *, char *,
      int, int, int, XSizeHints *) ;
extern "C" void XSelectInput(Display *, Window, int) ;
extern "C" void XSync(Display *, int) ;
extern "C" void XMapWindow(Display *, Window) ;
extern "C" void XNextEvent(Display *, XEvent *) ;
extern "C" void XDrawLine(Display *, Window, GC, int, int, int, int) ;
extern "C" void XDrawArc(Display *, Window, GC,
                                    int, int, int, int, int, int) ;
extern "C" void XDrawRectangle(Display *, Window, GC, int, int, int, int) ;
extern "C" void XFlush(Display *) ;
extern "C" int XAllocNamedColor(Display *, Colormap, char *,
                                    XColor *, XColor *) ;
extern "C" void XSetForeground(Display *, GC, int) ;
extern "C" void XSetFont(Display *, GC, Font) ;
extern "C" void XDrawString(Display *, Window, GC, int, int,
                                    const char *, int) ;
extern "C" XFontStruct *XLoadQueryFont(Display *, const char *) ;
void error(const char *) ;
#else
#include <X11/Xlib.h>
#include <X11/Xutil.h>
void error() ;
#endif
static int whitePixel, blackPixel ;
static Display *display ;
static GC gc ;
static int colortable[4] ;
XFontStruct *font ;
init_X(VOID) {
   XColor xcolor, dummy ;
   if (display == 0) {
      if ((display = XOpenDisplay(0))==0)
         error("! can't open display") ;
      whitePixel = WhitePixel(display, DefaultScreen(display)) ;
      blackPixel = BlackPixel(display, DefaultScreen(display)) ;
      gc = DefaultGC(display, DefaultScreen(display)) ;
      colortable[0] = whitePixel ;
      colortable[1] = blackPixel ;
      if (XAllocNamedColor(display, DefaultColormap(display,
             DefaultScreen(display)), "red", &xcolor, &dummy))
         colortable[2] = xcolor.pixel ;
      else
         colortable[2] = blackPixel ;
      if (XAllocNamedColor(display, DefaultColormap(display,
             DefaultScreen(display)), "blue", &xcolor, &dummy))
         colortable[3] = xcolor.pixel ;
      else
         colortable[3] = blackPixel ;
      font = XLoadQueryFont(display, "6x10") ;
      if (font == 0)
         error("! no such font") ;
      XSetFont(display, gc, font->fid) ;
   }
}
static int hoff, voff ;
static Window window ;
#ifdef CPLUS
static int S_open_window(int width, int height)
#else
static int S_open_window(width, height)
int width, height ;
#endif
{
   XSizeHints xsizehints ;
   XSetWindowAttributes xsetwindowattributes ;
   XEvent event ;
   int x, y ;

   hoff = 5 ;
   voff = 20 ;
   width += 10 ;
   height += 30 ;
   if (width > 1280)
      x = (1280 - width) / 2 ;
   else
      x = 20 ;
   if (height > 1024)
      y = (1024 - height) / 2 ;
   else
      y = 20 ;
   init_X() ;
   xsetwindowattributes.background_pixel = whitePixel ;
   xsetwindowattributes.border_pixel = blackPixel ;
   xsetwindowattributes.backing_store = Always ;
   window = XCreateWindow(display, DefaultRootWindow(display), x, y,
      width, height, 2, DefaultDepth(display, DefaultScreen(display)),
      InputOutput, DefaultVisual(display, DefaultScreen(display)),
      CWBackingStore|CWBackPixel|CWBorderPixel, &xsetwindowattributes) ;
   if (window == 0)
      error("! couldn't open window") ;
   xsizehints.x = x ;
   xsizehints.y = y ;
   xsizehints.width = width ;
   xsizehints.height = height ;
   xsizehints.flags = USSize | USPosition ;
   XSetStandardProperties(display, window, progname, progname, None,
      0, 0, &xsizehints) ;
   XSelectInput(display, window, StructureNotifyMask) ;
   XSync(display, 1) ;
   XMapWindow(display, window) ;
   XNextEvent(display, &event) ;
   return (int)window ;
}
#ifdef CPLUS
void S_close_window(int w)
#else
void S_close_window(w)
int w ;
#endif
{
   if (w)
      XDestroyWindow(display, w) ;
   XFlush(display) ;
}

#ifdef CPLUS
static void S_draw_line(int x1, int y1, int x2, int y2)
#else
static void S_draw_line(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   XDrawLine(display, window, gc, x1+hoff, y1+voff, x2+hoff, y2+voff) ;
}

#ifdef CPLUS
static void S_draw_circle(int x1, int y1, int x2, int y2)
#else
static void S_draw_circle(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   XDrawArc(display, window, gc, x1+hoff, y1+voff, x2, y2, 0, 360 * 64) ;
}

#ifdef CPLUS
static void S_draw_box(int x1, int y1, int x2, int y2)
#else
static void S_draw_box(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   XDrawRectangle(display, window, gc, x1+hoff, y1+voff, x2, y2) ;
}

static void S_flush_window(VOID) {
   XFlush(display) ;
}

#ifdef CPLUS
static void S_set_color(int c)
#else
static void S_set_color(c)
int c ;
#endif
{
   XSetForeground(display, gc, colortable[c]) ;
}

#ifdef CPLUS
static void S_label(int x, int y, const char *s)
#else
static void S_label(x, y, s)
int x, y ;
const char *s ;
#endif
{
   int len = strlen(s) ;
   x -= len * 3 ;
   y += 5 ;
   XDrawString(display, window, gc, x+hoff, y+voff, s, len) ;
}

#ifdef CPLUS
static int S_string_width(const char *s)
#else
static int S_string_width(s)
char *s ;
#endif
{
   return strlen(s) * 6 ;
}

static int S_string_height(VOID)
{
   return 10 ;
}
#endif
#ifdef AMIGA
#include "intuition/intuition.h"
#include "graphics/gfxbase.h"
static struct Window *window ;
struct IntuitionBase *IntuitionBase ;
struct GfxBase *GfxBase ;
static struct NewWindow newwindow = {
   0, 0, 0, 0, -1, -1, CLOSEWINDOW, WINDOWDEPTH | WINDOWCLOSE | WINDOWDRAG
   | SMART_REFRESH | NOCAREREFRESH, 0, 0,
   (UBYTE *)"Layout Display Window", 0, 0, 60, 25, -1, -1, WBENCHSCREEN } ;
void cleanup() {
   if (window)
      CloseWindow(window) ;
   if (IntuitionBase)
      CloseLibrary(IntuitionBase) ;
   if (GfxBase)
      CloseLibrary(GfxBase) ;
   exit(0) ;
}
static int hoff = 5 ;
static int voff = 20 ;
static int S_open_window(width, height)
int width, height ;
{
   extern void *OpenWindow() ;

   hoff = 5 ;
   voff = 20 ;
   width += 10 ;
   height += 30 ;
   if (width < 100) {
      hoff = (100 - width) / 2 + 5 ;
      width = 100 ;
   }
   if (IntuitionBase == 0) {
      if ((IntuitionBase =
         (struct IntuitionBase *)OpenLibrary("intuition.library",0L))==0 ||
          (GfxBase = 
               (struct GfxBase *)OpenLibrary("graphics.library",0L))==0)
         error("! no libraries") ;
   }
   if (width > GfxBase->NormalDisplayColumns)
      width = GfxBase->NormalDisplayColumns ;
   if (height > 2 * GfxBase->NormalDisplayRows)
      height = 2 * GfxBase->NormalDisplayRows ;
   newwindow.Height = height ;
   newwindow.Width = width ;
   if ((window = OpenWindow(&newwindow))==0)
      error("! can't get window.") ;
   return (int)window ;
}

static void S_close_window(w)
int w ;
{
   if (w)
      CloseWindow(w) ;
}

static void S_draw_line(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
{
   Move(window->RPort, x1+hoff, y1+voff) ;
   Draw(window->RPort, x2+hoff, y2+voff) ;
}

static void S_draw_box(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
{
   Move(window->RPort, x1+hoff, y1+voff) ;
   Draw(window->RPort, x1+hoff, y1+voff+y2) ;
   Draw(window->RPort, x1+hoff+x2, y1+voff+y2) ;
   Draw(window->RPort, x1+hoff+x2, y1+voff) ;
   Draw(window->RPort, x1+hoff, y1+voff) ;
}

struct IntuiText itext = { 1, 0, JAM1, 0, 0, NULL, 0 } ;

static void S_label(x, y, s)
int x, y ;
char *s ;
{
   int len = strlen(s) ;
   itext.IText = (UBYTE *)s ;
   PrintIText(window->RPort, &itext, x-4*len+hoff, y-5+voff) ;
   
}

static void S_draw_circle(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
{
   DrawEllipse(window->RPort, x1+x2/2+hoff, y1+y2/2+voff, x2/2, y2/2) ;
}

static void S_set_color(n)
int n ;
{
   SetAPen(window->RPort, n) ;
}
static void S_flush_window() {}

static int S_string_width(s)
char *s ;
{
   return strlen(s) * 8 ;
}

static int S_string_height()
{
   return 11 ;
}
#endif
/*
/   Now the PostScript output routines.
/*/
static short helvwidth[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   278, 278, 355, 556, 556, 889, 667, 222,
   333, 333, 389, 584, 278, 333, 278, 278,
   556, 556, 556, 556, 556, 556, 556, 556,
   556, 556, 278, 278, 584, 584, 584, 556,
   1015, 667, 667, 722, 722, 667, 611, 778,
   722, 278, 500, 667, 556, 833, 722, 778,
   667, 778, 722, 667, 611, 722, 667, 944,
   667, 667, 611, 278, 278, 278, 469, 556,
   222, 556, 556, 500, 556, 556, 278, 556,
   556, 222, 222, 500, 222, 833, 556, 556,
   556, 556, 333, 500, 278, 556, 500, 722,
   500, 500, 500, 334, 260, 334, 584, 0 } ;

FILE *psfile ;

#ifdef CPLUS
static void P_close_window(int w) {}
#else
static void P_close_window(w)
int w ;
{}
#endif

int yheight ;

static int y1cache, y2cache, colorcache ;

#ifdef CPLUS
static int P_open_window(int x, int y)
#else
static int P_open_window(x, y)
int x, y ;
#endif
{
   yheight = y ;
   fprintf(psfile, "%%!PS\n%%%%BoundingBox: 72 72 %d %d\n", x+72, y+72) ;
   fprintf(psfile, "/X{exch def}def\n") ;
   fprintf(psfile, "{x1 y1 moveto x2 y2 lineto stroke}/S X\n") ;
   fprintf(psfile, "{G S}/L X{/x2 X /x1 X S}/l X\n") ;
   fprintf(psfile, "{/y2 X /x2 X /y1 X /x1 X}/G X\n") ;
   fprintf(psfile, "{G x1 y1 moveto 0 y2 neg rlineto x2 0 rlineto\n") ;
   fprintf(psfile, " 0 y2 rlineto closepath stroke}/B X\n") ;
   fprintf(psfile, "{3 eq{0.4 setgray}{0 setgray}ifelse}/C X\n") ;
   fprintf(psfile, "{G x1 x2 2 div add y1 y2 2 div sub\n") ;
   fprintf(psfile, " x2 2 div 0 360 newpath arc stroke}/A X\n") ;
   fprintf(psfile, "{moveto show}/T X\n") ;
   fprintf(psfile, "%%%%EndProlog\n%%%%Page: 1 1\n") ;
   fprintf(psfile, "72 72 translate\n") ;
   fprintf(psfile, "/Helvetica findfont 10 scalefont setfont\n") ;
   colorcache = -1 ;
   y1cache = -1 ;
   y2cache = -1 ;
   return 0 ;
}

#ifdef CPLUS
static void P_draw_line(int x1, int y1, int x2, int y2)
#else
static void P_draw_line(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   if (y1 == y1cache && y2 == y2cache)
      fprintf(psfile, "%d %d l\n", x1, x2) ;
   else {
      fprintf(psfile, "%d %d %d %d L\n", x1, yheight-y1, x2, yheight-y2) ;
      y1cache = y1 ;
      y2cache = y2 ;
   }
}

#ifdef CPLUS
static void P_draw_box(int x1, int y1, int x2, int y2)
#else
static void P_draw_box(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   fprintf(psfile, "%d %d %d %d B\n", x1, yheight-y1, x2, y2) ;
   y1cache = -1 ;
   y2cache = -1 ;
}

#ifdef CPLUS
static void P_set_color(int n)
#else
static void P_set_color(n)
int n ;
#endif
{
   if (colorcache != n) {
      fprintf(psfile, "%d C ", n) ;
      colorcache = n ;
   }
}

static void P_flush_window(VOID) {
   fprintf(psfile, "showpage\n") ;
   if (psfile != stdout)
      fclose(psfile) ;
   psfile = stdout ;
}


#ifdef CPLUS
static void P_draw_circle(int x1, int y1, int x2, int y2)
#else
static void P_draw_circle(x1, y1, x2, y2)
int x1, y1, x2, y2 ;
#endif
{
   fprintf(psfile, "%d %d %d %d A\n", x1, yheight-y1, x2, y2) ;
   y1cache = -1 ;
   y2cache = -1 ;
}

#ifdef CPLUS
static int P_string_width(const char *s)
#else
static int P_string_width(s)
char *s ;
#endif
{
   int sum = 0 ;
   while (*s)
      sum += helvwidth[*s++] ;
   return (sum + 50) / 100 ;
}

static int P_string_height(VOID) {
   return 10 ;
}

#ifdef CPLUS
static void P_label(int x, int y, const char *s)
#else
static void P_label(x, y, s)
int x, y ;
char *s ;
#endif
{
   fprintf(psfile, "(%s) %d %d T\n", s, x-P_string_width(s)/2,
                                        yheight-y-P_string_height()/2+1) ;
}
/*
/   Now the selection between the two possibilities.
/*/
#include "display.h"
struct display_routines P_display = {
   P_close_window, P_open_window, P_draw_line, P_draw_box, P_set_color,
   P_flush_window, P_draw_circle, P_label, P_string_width, P_string_height} ;
struct display_routines S_display = {
   S_close_window, S_open_window, S_draw_line, S_draw_box, S_set_color,
   S_flush_window, S_draw_circle, S_label, S_string_width, S_string_height} ;
struct display_routines *cur_display = &S_display ;
