;
;   Some macros to make things easier.
;
(defmacro (low? sig)
   `(= (value ,sig) 0))
;
(defmacro (high? sig)
   `(<> (value ,sig) 0))
;
(define (return) ())
;
(defmacro (wait-high sig)
   `(let loop ()
      (if (high? ,sig)
          (return)
          (loop))))
;
(defmacro (wait-low sig)
   `(let loop ()
      (if (low? ,sig)
          (return)
          (loop))))
;
(defmacro (forever . whatever)
   `(let loop ()
      ,(cons 'begin whatever)
      (loop)))
;
(defmacro dont-care ())
;
(defmacro w 24)
;
;   This is an example model of a differentiator.
;
(define (read stb:wire:in ack:wire:out bus:wire:w:in key:key)
   (model
      (forever
         (set! ack 0)
         (wait-high stb)
         (out key (value bus))
         (set! ack 1)
         (wait-low stb))))
;
(define (write stb:wire:out ack:wire:in bus:wire:w:out key:key)
   (model
      (forever
         (set! stb 0)
         (wait-low ack)
         (set! bus (in key))
         (set! stb 1)
         (wait-high ack)
         (set! bus dont-care))))
;
(define (differ key1 key2)
   (model
      (let loop ((sum:value:w 0))
xc         (let ((newval:value:w (in key1)))
            (out key2 (- newval sum))
            (loop newval)))))
;
(define (subcircuit instb:wire:in outstb:wire:out inack:wire:in outack:wire:out
                 inbus:wire:w:in outbus:wire:w:out)
   (let ((key1 (make-key))
         (key2 (make-key)))
      (par
         (lambda () (read instb outack inbus key1))
         (lambda () (write outstb inack outbus key2))
         (lambda () (differ key1 key2)))))
;
;   Some simulation stuff.
;
(define (prime-aux a b)
   (if (> (* b b) a) #t
       (if (= a (* b (/ a b))) #f
           (prime-aux a (+ b 2)))))
;
(define (prime a)
   (prime-aux a 3))
;
(define (lop n key)
   (if (prime n)
       (out key n))
   (lop (+ n 2) key))
;
(define (stim key)
   (out key 2)
   (lop 3 key))
;
;   And output stuff.
;
(define (record key)
   (display (in key) " " (obactive) "\n")
   (record key))
;
;   Finally, the main circuit.
;
(define (circuit)
   (let ((stb1:wire 0) (stb2:wire 0) (stb3:wire 0)
         (ack1:wire 0) (ack2:wire 0) (ack3:wire 0)
         (bus1:wire:w) (bus2:wire:w) (bus3:wire:w)
         (key1 (make-key)) (key2 (make-key)))
      (par
         (lambda () (stim key1))
         (lambda () (write stb1 ack1 bus1 key1))
         (lambda () (subcircuit stb1 stb2 ack2 ack1 bus1 bus2))
         (lambda () (subcircuit stb2 stb3 ack3 ack2 bus2 bus3))
         (lambda () (read stb3 ack3 bus3 key2))
         (lambda () (record key2)))))
