;
;   A for loop.
;
(defmacro (for var init test next body)
   `(let loop ((,var ,init))
      (if ,test
          (begin
             ,body
             (loop ,next))
          #t)))
;
;   A for loop for an integer sequence.
;
(defmacro (forint var init max body)
   `(for ,var ,init (< ,var ,max) (+ ,var 1) ,body))
;
;   A dolist macro.
;
(defmacro (dolist var list body)
   `(let loop ((llop ,list))
      (if llop
         (let ((,var (car llop)))
            ,body
            (loop (cdr llop))))))
;
;   And, done by if's.
;
(defmacro (and . b)
   (if b
       (let ((first (car b))
             (rest (cdr b)))
          (if rest
             (if (atom? first)
                `(if ,first (and . ,rest) ())
                (let ((g (gensym)))
                   `(let ((,g ,first))
                       (if ,g (and . ,rest) ()))))
              first))
          '#t))
;
;   Or, done by if's.
;
(defmacro (or . b)
   (if b
       (let ((first (car b))
             (rest (cdr b)))
          (if rest
             (if (atom? first)
                `(if ,first ,first (or . ,rest))
                (let ((g (gensym)))
                   `(let ((,g ,first))
                       (if ,g ,g (or . ,rest)))))
              first))
          '()))
;
(define (list->array lst)
   (let* ((n (length lst))
          (a (make-array n)))
      (let loop ((i 0) (ls lst))
         (if ls
            (begin
               (aset a i (car ls))
               (loop (+ i 1) (cdr ls)))
            a))))
;
(define (gensym-gen)
   (let ((gctr 0))
      (lambda ()
         (set! gctr (+ gctr 1))
         (str->sym (strcat "%" gctr)))))
;
(define gensym (gensym-gen))
;
(define (min a b)
   (if (< a b) a b))
;
(define (max a b)
   (if (> a b) a b))
;
(display "\n")
