;;;
;
;   For convenience . . .
;
(defmacro (wait-high sig)
   `(let loop ()
       (if (value ,sig)
           (return)
           (loop))))
;
(defmacro (wait-low sig)
   `(let loop ()
       (if (value ,sig)
           (loop)
           (return))))
;
(defmacro (lower sig)
   `(emit ,sig 0))
;
(defmacro (raise sig)
   `(emit ,sig 1))
;
(defmacro (forever . whatever)
   `(let loop ()
      (begin . ,whatever)
      (loop)))
;;;
;   First, our global wires
;
;   On the input side, we have a strobe, data bus, and acknowledge.
;
;   These don't have to be global; they can be instantiated by a
;   process and passed in to `make-process'.
;
(define strobe_in:in:wire 0)
(define ack_in:out:wire 0)
(define data_in:in:wire:32 0)
;
;   Similarly on the output.
;
(define strobe_out:out:wire 0)
(define ack_out:in:wire 0)
(define data_out:out:wire:32 0)
;
;   We've got to make processes for each of these.
;
(define (read:noreturn)
   (forever
      (lower ack_in)
      (wait-high strobe_in)
      (rendezvous (get-data:value:32)
         (return (value data_in)))
      (raise ack_in)
      (wait-low strobe_in)))
;
(define (write:noreturn)
   (forever
      (lower strobe_out)
      (wait-low ack_out)
      (rendezvous (put-data:void val:value:32)
         (return)
         (emit data_out val))
      (raise strobe_out)
      (wait-high ack_out)))
;
;   Our computation:  mgcd.
;
(define (mgcd:value:32 a:value:32 b:value:32)
   (if (= a b) a
       (if (< a b)
           (mgcd a (- b a))
           (mgcd b a))))
;
;   Our final circuit.
;
(define (main:noreturn)
   (let ((read-proc (read))
         (write-proc (write))
         (forever
            (let ((a:value:32 (read-proc get-data)))
               (write-proc put-data (mgcd a (read-proc get-data))))))))
