;
;   Floyd's all pairs shortest path algorithm.
;
(define *n* 0)
;
;   This is a version of Floyd's algorithm where you pass in
;   your < and + routines.  Thank God for Scheme; the syntax
;   and everything remains the same.
;
(define (floyd ar n < +)
   (for i 0 (< i n) (+ i 1)
      (for j 0 (< j n) (+ j 1)
         (for k 0 (< k n) (+ k 1)
            (let ((sum (+ (aref ar i k) (aref ar k j))))
               (if (< sum (aref ar i j))
                  (aset ar i j sum)))))))
;
(define infty 'infty)
(define sprng 'sprng)
;
;   Here are some arithmetic operations for our extended numbers.
;
;   The only values represented as cons cells are spring values.
;   This may change, but probably not.
;
(define (spring? a)
   (pair? a))
;
;   This function takes an integer and springs it.
;
(define (spring a)
   (if (spring? a)
      a
      (cons a sprng)))
;
;   This value strips the spring off a number if it has any.
;
(define (unspring a)
   (if (spring? a)
      (car a)
      a))
;
;   This adds two times.
;
(define (time+ a b)
   (if (or (eq a infty) (eq b infty)) infty
      (if (or (spring? a) (spring? b))
         (spring (+ (unspring a) (unspring b)))
         (+ a b))))
;
;   And this compares two times.
;
(define (time< a b)
   (if (eq a infty) #f
      (if (eq b infty) #t
         (if (spring? a)
            (if (spring? b)
               (> (unspring a) (unspring b))
               #f)
            (if (spring? b)
               #t
               (< a b))))))
;
(define (init n)
   (define *n* n)
   (define a (make-array n n))
   (for i 0 (< i n) (+ i 1)
      (for j 0 (< j n) (+ j 1)
         (aset a i j infty))))
;
(define (setdiagonal v)
   (for i 0 (< i *n*) (+ i 1)
      (aset a i i v)))
;
(define (delay x y min max)
   (aset a x y max)
   (aset a y x (- 0 min)))
;
;   To add an unbounded.
;
(define (unbounded x y)
   (delay x y 0 infty))
;
;   To add a spring.
;
(define (springed x y)
   (delay x y 0 (spring 0)))
;
(define (domats)
   (nicedisplay a)
   (setdiagonal infty)
   (floyd a *n* time< time+)
   (nicedisplay a))
;
(define (showdelay x y)
   (display x " " y " " (- 0 (aref a y x)) " " (aref a x y) "\n"))
;
;   We want to be able to nicely display an array.  We do this by
;   building strings for each element, and creating a new array
;   with the strings.  We assume it is a 2d array.
;
;   We also do pagination tricks; count the length of columns and
;   the like.
;
(define (nicedisplay a)
   (let ((new-array (make-array *n* *n*))
         (col-size (make-array *n*))
         (n *n*))
      (for i 0 (< i n) (+ i 1)
         (aset col-size i 0))
      (for i 0 (< i n) (+ i 1)
         (for j 0 (< j n) (+ j 1)
            (let ((el (aref a i j)))
                (aset new-array i j
                   (if (spring? el)
                      (strcat "S" (unspring el))
                      (if (eq el infty)
                         "Infty"
                         (strcat el))))
                (aset col-size j (max (aref col-size j)
                                      (length (aref new-array i j)))))))
      (for i 0 (< i n) (+ i 1)
         (for j 0 (< j n) (+ j 1)
            (let loop ((istr (aref new-array i j)))
               (if (> (aref col-size j) (length istr))
                  (loop (strcat " " istr))
                  (aset new-array i j istr)))))
      (display new-array "\n")))

