;;;pfilt.hs
;
;   This is a parallel prime algorithm that works by spawning filters,
;   each of which removes numbers that are multiples of them.
;
(define (filter inkey outkey primeinc accum)
   (let loop ((temp (in inkey)) (accum accum))
      (if (< temp accum)
          (begin
             (out outkey temp)
             (loop (in inkey) accum))
          (if (> temp accum)
              (loop temp (+ accum primeinc))
              (loop (in inkey) (+ accum primeinc))))))
;
;   We don't want the filters to kick in too fast, though.  This allows
;   a filter to bypass the rest of the filters until its time has come.
;
(define (first-filter inkey outkey fastkey primeinc accum)
   (let loop ((temp (in inkey)))
      (if (< temp accum) 
         (begin (out fastkey temp)
                (loop (in inkey)))
         (filter inkey outkey primeinc (+ accum primeinc)))))
;
;   Our odd generator.
;
(define (oddgen key n)
   (out key n)
   (oddgen key (+ n 2)))
;
;   Our routine that builds the filter chain.
;
(define (filter-chain inkey chain-key)
   (let ((new-num (in inkey))
         (new-key (make-key)))
      (display new-num " ")
      (par
         (lambda () (first-filter chain-key new-key inkey
                                         (* 2 new-num) (* new-num new-num)))
         (lambda () (filter-chain inkey new-key)))))
;
;   Finally, we run.
;
(let ((new-key (make-key))
      (forever-key (make-key)))
   (out forever-key 3)
   (par (lambda () (oddgen new-key 5))
        (lambda () (filter-chain forever-key new-key))))
