;   This is my regression test file for the hws system.  First, we
;   test the scheme parts of the system.
;   Basic operation by fib.

(display "The numbers at the top and bottom of this output should match\n")

(display (obactive) "\n")

(define (newline) (display "\n"))

(define (fib a)
   (if (< a 2) 1
       (+ (fib (- a 1)) (fib (- a 2)))))

(display "This should be 89: " (fib 10) "\n")

;  Now we test whether or not we are getting lambdas.

(define (a n)
   (cons (lambda () n)
         (lambda (c) (set! n c) n)))

(define foo (a 10))
(define bar (a 20))
(display "This should be 10: " ((car foo)) " and 20: " ((car bar)) "\n")
((cdr foo) 30)
((cdr bar) 40)
(display "This should be 30: " ((car foo)) " and 40: " ((car bar)) "\n")

;   More nested scopes.

(define (a n)
   (if (< n 0)
       '()
       (cons (lambda () n)
             (a (- n 1)))))

(define foo (a 10))

(define (doem n l)
   (if (< n 0)
       '()
       (cons ((car l))
             (doem (- n 1) (cdr l)))))

(display "This should be 10..0: " (doem 10 foo) "\n")

;   Now we test the `let' statement; does it work?

(define (foo a b)
   (let ((a (+ a a))
         (b (+ a b)))
      (+ a (* b 1000))))

(display "This should be 7006: " (foo 3 4) "\n")

;   Now we test the `let*' statement; does it work?

(define (foo a b)
   (let* ((a (+ a a))
          (b (+ a b)))
      (+ a (* b 1000))))

(display "This should be 10006: " (foo 3 4) "\n")

;   Tail recursion:  find the fiftieth prime.  (This will blow out the
;   stack on an Amiga unless tail recursion works.)

(define (prime-aux a b)
   (if (> (* b b) a) #t
       (if (= a (* b (/ a b))) #f
           (prime-aux a (+ b 2)))))

(define (prime a)
   (prime-aux a 3))

(define (lop n)
   (if (prime n)
       (begin
          (display n)
          (lop (+ n 2)))
       (lop (+ n 2))))

(define (nth n)
   (if (<= n 1) 2
       (nth-aux (- n 1) 3)))

(define (nth-aux n m)
   (if (prime m)
       (if (<= n 1) m
           (nth-aux (- n 1) (+ m 2)))
       (nth-aux n (+ m 2))))

(display "This should be 29: " (nth 10) "\n")

;   Now we test the named let.

(define (iota n)
   (let loop ((n n) (res ()))
       (if (<= n 0) res
           (loop (- n 1) (loop (- n 1) (cons n res))))))

(display "This should be (1 1 2 1 1 2 3 1 1 2 1 1 2 3 4): " (iota 4) "\n")

;   Now we test nested named lets.

(define (test n)
   (let loop ((y 0) (s 0))
      (if (< y n)
          (loop (+ y 1)
                (+ s (let loop ((x 0) (s 0))
                         (if (< x n)
                             (loop (+ x 1) (+ s (* x y)))
                             s))))
          s)))

(display "This should be 100: " (test 5) "\n")

;   Time to test typing.  We need to test it for lambda binding, for let
;   binding, for named lets, and for define/set!.

(define (8+:8 a:8 b:unsigned:8)
   (* a b))

(display "This should be -1518: " (8+ 234 3141) "\n")

(define b:u:3 -1)
(display "This should be 7: " b "\n")

(define (loopmul)
   (let foo ((a:unsigned:8 65537) (b:unsigned:16 65536))
      (if (= a 0) b
          (foo (+ a 1) (+ b (* a a))))))

(display "This should be 54656: " (loopmul) "\n")

;   Can we parse hexadecimal numbers?

(display "This should be (1 10 43981 -1): "
         '(0x1 0xa 0xaBcD 0xffffffff) "\n")

;   We had a problem with immediate lambdas; let's test them.

(display "This should be 12: "
            ((let ((n 12)) (lambda () n))) "\n")

(display "This should be 14: "
            (((lambda (n) (lambda () n)) 14)) "\n")

(display "This should be (a . c): "
            ((((lambda (n) (lambda (m) (lambda () (cons m n)))) 'c) 'a)) "\n")

;   Now we test wires, which are `by reference'.

(display-type a)
(define a:wire:3 0)
(display-type a)

(define (b c:wire:3 v)
   (set! c v))

(b a 4)

(display "This should be W-4: " a "\n")

(b a 5)

(display "This should be W-3: " a "\n")

;   We have to test local allocation of wires.

(define (local-test foo)
   (let ((my-wire:wire:3 0))
      (b my-wire foo)
      (value my-wire)))

(display "This should be 1: " (local-test 9) "\n")

;   Test basic operation of in/out.

(define key1 (make-key))

(define key2 (make-key))

(out key1 9394)

(out key2 300)

(display "This should be 9094: " (- (in key1) (in key2)) "\n")

;   Test basic operation of par

(define (sender key n)
   (let foo ((i 1) (n n))
      (out key i)
      (if (< n 2)
          ()
          (foo (+ i i) (- n 1)))))

(define (receiver key n)
   (let foo ((s 0) (n n))
      (if (< n 1)
          s
          (foo (+ s (in key)) (- n 1)))))

(define rval 0)

(par (lambda () (sender key1 7))
     (lambda () (set! rval (receiver key1 7))))

(display "This should be 127: " rval "\n")

;   Now test memory leaks.

(clear)

(display (obactive) "\n")
