;
;   Analyzes timed petri nets.  Not memory efficient; intended to
;   show how the algorithm works.
;
(define MAXN 20)
(define *bounds* (make-array MAXN MAXN))
(define *skew* (make-array MAXN MAXN))
(define *preds* (make-array MAXN))
;
(define *n* 0)
;
;   The symbolic math stuff.
;
(define infty 'infty)
(define -infty '-infty)
;
(define (sym-add a b)
   (if (eq a infty)
      (if (eq b -infty)
         (error "! can't add infinity to -infinity")
         a)
      (if (eq b infty)
         (if (eq a -infty)
            (error "! can't add infinity to -infinity")
            b)
         (if (or (eq a -infty) (eq b -infty)) -infty
            (+ a b)))))
;
(define (sym-max a b)
   (if (or (eq a infty) (eq b infty)) infty
      (if (eq a -infty) b
         (if (eq b -infty) a
            (if (< a b) b a)))))
;
(define (sym-min a b)
   (if (or (eq a -infty) (eq b -infty)) -infty
      (if (eq a infty) b
         (if (eq b infty) a
            (if (< a b) a b)))))
;
;   Now the code to add a node.  Strictly housekeeping.
;
(define (add-node arc-list)
   (aset *preds* *n* '())
   (forint i 0 *n*
      (begin
         (aset *skew* i *n* -infty)
         (aset *skew* *n* i infty)))
   (aset *skew* *n* *n* 0)
   (dolist arc arc-list
      (let ((pred (car arc))
            (min (car (cdr arc)))
            (max (car (cdr (cdr arc)))))
         (aset *preds* *n* (cons pred (aref *preds* *n*)))
         (aset *bounds* pred *n* max)
         (aset *bounds* *n* pred (- 0 min))))
   (find-skew-for-node-*n*)
   (display-arrays)
   (set! *n* (+ *n* 1))
   "")
;
;   Next, the actual meat of the algorithm.
;
(define (find-skew-for-node-*n*)
   (let ((predlist (aref *preds* *n*)))
      ;
      ;   This first section calculates the maximum from any node
      ;   to the one we are adding by assuming maximal delay at each
      ;   point.
      ;
      (forint i 0 *n*
         (if predlist
            (dolist pred predlist
               (aset *skew* i *n*
                   (sym-max (aref *skew* i *n*)
                            (sym-add (aref *skew* i pred)
                                     (aref *bounds* pred *n*)))))
            (aset *skew* i *n* infty)))
      ;
      ;   This section calculates the minimum delay from any node
      ;   to the new one, neglecting `and' conditions and assuming
      ;   minimum delay.
      ;
      (dolist pred predlist
         (forint i 0 *n*
            (aset *skew* *n* i
               (sym-min (aref *skew* *n* i)
                        (sym-add (aref *skew* pred i)
                                 (aref *bounds* *n* pred))))))
      ;
      ;   Finally, here we tighten the lower bounds by factoring
      ;   in the skews among the predecessors of our new node.
      ;
      (forint i 1 *n*
         (let ((max-delta -infty))
            (dolist pred (aref *preds* i)
               (set! max-delta (sym-max max-delta
                                        (sym-add (aref *skew* *n* pred)
                                                 (aref *bounds* pred i)))))
            (aset *skew* *n* i
                (sym-min (aref *skew* *n* i) max-delta))))))
;
;   And the reporting routine.
;
(define (display-arrays)
   (display "Now we have " (+ *n* 1) " nodes.\n")
   (display "(-min,max) to node " *n* "\n")
   (forint i 0 *n*
      (display i ":  (" (aref *skew* *n* i) "," (aref *skew* i *n*) ")\n")))
;
